from __future__ import annotations

import logging

import wx

from aio_wx_widgets import type_annotations as T
from aio_wx_widgets.binding import Binding
from aio_wx_widgets.panels.panel import SimplePanel
from aio_wx_widgets.sizers import AlignHorizontal
from aio_wx_widgets.widgets.button import AioButton
from aio_wx_widgets.widgets.grid import Grid, VERTICAL
from aio_wx_widgets.widgets.select import Select
from aio_wx_widgets.widgets.text import Text
from aio_wx_widgets.widgets.text_entry import IntEntry

_LOGGER = logging.getLogger(__name__)


class ViewTwo(SimplePanel):
    """Second view."""

    def __init__(self, parent, controller):
        self._controller = controller
        super().__init__(parent)
        self._text = Text("This is text that changes as a result of an event happening")
        self._controller.add_to_log.subscribe(self._add_to_log)
        self._callback_text = Text("Nothing selected yet")
        self._bound_selected_item = Text("Nothing selected yet")

    @property
    def controller(self):
        return self._controller

    def populate(self):
        self.add(
            IntEntry(binding=Binding(self._controller, "value_1")),
            margin=(10, 10, 5, 20),
        )
        self.add(IntEntry(binding=Binding(self._controller, "value_1")))
        self.add(
            AioButton("test button with normal function as callback", self._press),
            margin=(10, 10, 5, 20),
        )
        self.add(self._text)

        # It is possible to add bare wx-python widgets.
        self.add(wx.StaticLine(parent=self.ui_item), create=False)

        with self.add(Grid()) as grd:
            grd.add(
                Text("This is a selection dropdown"),
                weight=4,
                align_horizontal=AlignHorizontal.right,
            )
            with grd.add(Grid(orientation=VERTICAL), weight=6) as grd:
                grd.add(
                    Select(
                        self.controller.choices,
                        on_select_callback=self._on_select_callback,
                        binding=self.bind("selected_item"),
                    ),
                    align_horizontal=AlignHorizontal.left,
                )
                grd.add(Text("Selection based on callback:"))
                grd.add(self._callback_text)
                grd.add(Text("Selection based on binding value:"))
                grd.add(
                    AioButton(
                        label="Display bound value", callback=self._show_bound_value
                    )
                )
                grd.add(self._bound_selected_item)

    def _show_bound_value(self, evt):
        self._bound_selected_item.set_text(self._controller.selected_item)

    def _on_select_callback(self, choice: T.Choice):
        self._callback_text.set_text(str(choice))

    def _press(self, evt):
        self._add_to_log("button pressed.")

    def _add_to_log(self, item: str):
        self._text.set_text(item)



import logging

from aio_wx_widgets.frame import DefaultFrame
from demo.model.demo_model import DemoModel
from demo.controller import controller_three
from demo.views.view_three import ViewThree

_LOGGER = logging.getLogger(__name__)


class Window(DefaultFrame):
    def __init__(self):
        super().__init__("third window")

        model = DemoModel()
        controller = controller_three.ControllerThree(model)
        view = ViewThree(self, controller)
        view.populate()



import asyncio
import logging

import pytest

from aio_wx_widgets.frame import DefaultFrame

_LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "kwargs,results",
    [
        ({"title": "a title"}, {"title": "a title", "width": 800, "height": 600}),
        (
            {"title": "a title", "size": (300, 200)},
            {"title": "a title", "width": 300, "height": 200},
        ),
    ],
)
def test_frame_initialize(wx_app, kwargs, results):
    """smoke test initializing the frame."""

    frame = DefaultFrame(**kwargs)

    assert frame.Title == results["title"]

    assert frame.Size.Width == results["width"]
    assert frame.Size.Height == results["height"]


@pytest.mark.asyncio
async def test_on_close_event(wx_app, mocker):
    """Test whether the _on_close callback is called when frame gets closed."""
    mocker.patch.object(DefaultFrame, "_on_close")
    frame = DefaultFrame(title="a title")
    frame.Close()
    await asyncio.sleep(0.1)
    frame._on_close.assert_called_once()



"""Property binding."""

import logging
from typing import NamedTuple, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from aio_wx_widgets.controller import BaseController

_LOGGER = logging.getLogger(__name__)

WATCHERS = "_watchers_"


class Binding(NamedTuple):
    """Binding data."""

    bound_object: "BaseController"
    bound_property: "str"


class Bindable:
    """Base class for binding a property to a widget.

    Make a widget inherit from this. The binding property modifies the controller bound
    property adding a watcher and emitting change events.
    """

    def __init__(self, binding: Optional["Binding"]):
        self._binding = binding

        # Flag to stop firing updates when a property value has changed.
        # Prevents infinite loop of updating the value and firing update events.
        self._fire_update_event = True

    def _make_binding(self):
        if self._binding is None:
            _LOGGER.debug("No binding defined.")
            return

        if WATCHERS not in self._binding.bound_object.__dict__:
            self._binding.bound_object.__dict__[WATCHERS] = {}
        if (
            self._binding.bound_property
            not in self._binding.bound_object.__dict__[WATCHERS]
        ):
            _LOGGER.debug("No watchers yet for %s", self._binding.bound_property)
            self._binding.bound_object.__dict__[WATCHERS][
                self._binding.bound_property
            ] = []
        self._binding.bound_object.__dict__[WATCHERS][
            self._binding.bound_property
        ].append(self._update)

        value = self.get_property_value()

        self._set_ui_value(value)

    def get_property_value(self):
        """Return the property value of the bound object.

        This would be the attribute defined in the controller.
        """
        return getattr(self._binding.bound_object, self._binding.bound_property)

    def _set_property_value(self, value):
        setattr(self._binding.bound_object, self._binding.bound_property, value)

    def _update(self, value):
        """Called when a property changes from outside this control.

        Update the control with the new value
        """
        _LOGGER.debug("Updating entry with new value.")
        self._fire_update_event = False
        self._set_ui_value(value)

    def _on_ui_change(self, *args, **kwargs):
        """Callback when UI widget value changes by user input."""
        if self._binding is None:
            return

        _LOGGER.debug("UI value changed. Trigger the binding.")
        if self._fire_update_event:
            _LOGGER.debug("Firing update")
            val = self._get_ui_value()
            self._set_property_value(val)
        self._fire_update_event = True

    def _get_ui_value(self):
        """Get the value from the ui widget."""
        raise NotImplementedError()

    def _set_ui_value(self, value):
        raise NotImplementedError()



import pytest
from wxasync import WxAsyncApp

from aio_wx_widgets.controller import BaseController


@pytest.fixture
def base_model():
    class Model:
        """A simple model"""

    return Model()


@pytest.fixture
def base_controller(base_model):
    return BaseController(model=base_model)


@pytest.fixture(scope="session")
def wx_app():
    app = WxAsyncApp()
    return app



import logging
from random import randint

from events import Events

from aio_wx_widgets.controller import BaseController
from demo.windows import second_window, third_window

_LOGGER = logging.getLogger(__name__)


class ControllerOne(BaseController):
    def __init__(self, model):
        self.value_1: int = 0
        self.a_string_value = "A certain string"
        super().__init__(model)

        self.add_to_log = Events()

    def _add_to_log(self, data: str):
        self.add_to_log.on_change(data)

    async def open_other_window(self):
        """Open a second window."""
        second_window.SecondWindow().Show()
        # _second_window.Show()
        _LOGGER.debug("Closed second window ?")

    async def open_third_window(self):
        third_window.Window().Show()

    async def set_value(self):
        val = randint(1, 100)
        _LOGGER.debug("Setting bound property to %s", val)
        self.value_1 = val



"""WX button."""

import logging
from asyncio import iscoroutinefunction
from typing import Callable, Any, Awaitable, Union

import wx
from wxasync import AsyncBind  # type: ignore

_LOGGER = logging.getLogger(__name__)


class AioButton:
    """Button widget.

    Args:
        label: button label.
        callback: a coroutine function or normal function. Called when pressed.
    """

    def __init__(
        self,
        label: str,
        callback: Union[Callable[[Any, Any], Awaitable], Callable[[Any, Any], None]],
    ):
        self.label = label
        self._call_back = callback

        self.ui_item = wx.Button()

    def __call__(self, parent):
        self.ui_item.Create(parent, label=self.label)
        if iscoroutinefunction(self._call_back):

            AsyncBind(wx.EVT_BUTTON, self._call_back, self.ui_item)
        else:
            self.ui_item.Bind(wx.EVT_BUTTON, self._call_back)
        return self



"""Text widgets."""

import logging
from typing import Optional

import wx

from aio_wx_widgets.binding import Bindable
from aio_wx_widgets.colors import GREEN
from aio_wx_widgets.const import is_debugging

_LOGGER = logging.getLogger(__name__)


class Text(Bindable):
    """Static text."""

    def __init__(
        self, text="", font_size=None, color: Optional[int] = None, binding=None
    ):
        """Init.

        Args:
            parent:
            text:
            font_size:
        """
        self._color = color
        self._text = text

        super().__init__(binding)

        self._font_size = None
        if font_size:
            self._font_size = wx.Font(wx.FontInfo(font_size))
        self.ui_item = wx.StaticText()

    def _set_ui_value(self, value):
        self.ui_item.SetLabelText(str(value))

    def _get_ui_value(self):
        """This is a one way binding. Not implementing this."""
        return None

    def __call__(self, parent):
        self.ui_item.Create(parent)
        if self._font_size:
            self.ui_item.SetFont(self._font_size)
        self.set_text(self._text, self._color)
        if is_debugging():
            self.ui_item.SetBackgroundColour(GREEN)
        self._make_binding()
        return self

    def set_text(self, text, color=None):
        """Set text."""
        self._text = str(text)
        self.ui_item.SetLabel(self._text)
        if color:
            self.ui_item.SetForegroundColour(color)



"""Datatypes used by widgets."""

import logging
from typing import NamedTuple, Any

_LOGGER = logging.getLogger(__name__)


class Choices(NamedTuple):
    """Choice item."""

    label: str
    value: Any



"""Sizer and Layout tooling."""

import collections
import logging
from enum import Enum
from typing import List, Tuple, Optional, Union

import wx

# pylint: disable=unused-import
from aio_wx_widgets import type_annotations as T
from aio_wx_widgets.binding import Binding

_LOGGER = logging.getLogger(__name__)


class AlignHorizontal(Enum):
    """Horizontal alignment options."""

    left = wx.ALIGN_LEFT
    right = wx.ALIGN_RIGHT
    center = wx.ALIGN_CENTER_HORIZONTAL


def _make_window(item: wx.Window, margins: List[Tuple[int, int]]):
    val = margins[0][1]
    for margin in margins[1:]:
        val = val | margin[1]

    new_sizer = wx.BoxSizer()
    margin_value = margins[0][0]
    new_sizer.Add(item, 1, val, margin_value)
    return new_sizer


def _margin_wrapper(item, margins: List[Tuple[int, int]]) -> wx.Window:
    if len(margins) == 0:
        return item

    first = margins.pop(0)

    values = [first]
    for idx, _ in enumerate(margins):
        if margins[idx][0] == first[0]:
            values.append(margins.pop(idx))

    item = _make_window(item, values)
    return _margin_wrapper(item, margins)


def _align_item(
    item: Union[wx.Window, wx.BoxSizer],
    current_sizer_orientation,
    alignment: Optional["AlignHorizontal"],
    current_layout: int,
) -> Tuple[Union[wx.Window, wx.BoxSizer], int]:

    if alignment is None:
        return item, current_layout | wx.EXPAND

    # wx_alignment = _get_wx_alignment(alignment)
    if not current_sizer_orientation == wx.HORIZONTAL:
        return item, current_layout | alignment.value

    sizer = wx.BoxSizer(orient=wx.VERTICAL)
    sizer.Add(item, 0, alignment.value, 0)
    return sizer, current_layout


def _add(
    item,
    parent,
    sizer,
    weight,
    margin,
    default_margin,
    create,
    align_horizontal: Optional[AlignHorizontal],
) -> object:
    if create:
        # this is an item which is part of the aio_wx_widgets family.
        # It is assumed it has the ui_item property.
        item = item(parent)
        ui_item = item.ui_item
    else:
        # this is assumed to be a "normal" wx-widgets item.
        ui_item = item
    if margin is None:
        margin = default_margin
    elif isinstance(margin, collections.Sequence):
        margins = [
            (margin[0], wx.LEFT),
            (margin[1], wx.RIGHT),
            (margin[2], wx.TOP),
            (margin[3], wx.BOTTOM),
        ]
        ui_item = _margin_wrapper(ui_item, margins)
        margin = 0

    layout = wx.ALL

    ui_item, layout = _align_item(ui_item, sizer.Orientation, align_horizontal, layout)

    sizer.Add(ui_item, weight, layout, margin)

    return item


class SizerMixin:
    """Mixin providing sizer functionality"""

    default_sizer_margin = 5

    def add(
        self,
        item,
        weight=0,
        layout=wx.EXPAND | wx.ALL,
        margin=None,
        create=True,
        align_horizontal: Optional[AlignHorizontal] = None,
    ):
        """Add an item to this panel

        Args:
            item: the item to be added.
            weight: The relative weight to other items in this container and the size
                it should take
            layout: The layout
            margin: None to use default. An int to use one margin for all sides.
                A tuple (left,right,top,bottom) for side specific margins.
            create: A ready made component can be added to.
        """
        try:
            parent = self._parent  # type: ignore
            if parent is None:
                raise AttributeError("parent cannot be none.")
        except AttributeError:
            parent = self.ui_item  # type: ignore

        return _add(
            item,
            parent,
            self._sizer,  # type: ignore
            weight,
            margin,
            SizerMixin.default_sizer_margin,
            create,
            align_horizontal=align_horizontal,
        )


class PanelMixin:
    """Mixin providing extra functionality."""

    def __init__(self):
        self.ui_item.Bind(wx.EVT_WINDOW_DESTROY, self._on_close)

    def _on_close(self, evt):
        _LOGGER.debug("Window destroyed.")
        self.controller.deactivate()
        evt.Skip()

    @property
    def controller(self) -> "T.BaseController":
        """Return the controller for this panel."""
        raise NotImplementedError()

    def bind(self, prop: str):
        """Bind a property to a bindable control.

        Args:
            prop: the attribute/property to bind.
        """
        return Binding(self.controller, prop)

    def populate(self):
        """Add and configure widgets here."""



"""Validators"""

# todo: Rethink this.

import logging
import string

import wx

_LOGGER = logging.getLogger(__name__)

# pylint: disable=invalid-name,no-self-use


class IntValidator(wx.Validator):
    """Validator."""

    def __init__(self):
        """Init."""
        wx.Validator.__init__(self)
        self.Bind(wx.EVT_CHAR, self._on_char)

    def Clone(self):
        """Clone."""
        return IntValidator()

    def _on_char(self, event):
        value = event.GetKeyCode()
        if value < wx.WXK_SPACE or value == wx.WXK_DELETE or value > 255:
            event.Skip()
            return
        if chr(value) in string.digits:
            event.Skip()
            return
        return

    def Validate(self, win):
        """Validate."""
        ctrl = self.GetWindow()
        value = ctrl.GetValue()
        try:
            int(value)
        except ValueError:
            return False
        return True



"""Group layout item. Draws a line around children and sets a title."""


import logging

import wx

from aio_wx_widgets.sizers import SizerMixin

_LOGGER = logging.getLogger(__name__)


class Group(SizerMixin):
    """A Group widget.

    +--<LABEL>-------------+
    |                      |
    |  - Widget 1          |
    |  - Widget 2          |
    |                      |
    +----------------------+

    """

    default_sizer_margin = 20

    def __init__(self, label, **kwargs):
        """Init.

        Args:
            label:
        """
        self._label = label
        self._extra_bottom_border = 10
        # if KEY_PLATFORM == KEY_LINUX:
        #     self._extra_bottom_border = 50
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.ui_item = wx.StaticBox()
        kwargs["sizer"] = self._sizer
        super().__init__()
        self._bottom_border = 1

    def __call__(self, parent):
        """Call this class as a function.

        Args:
            parent: The parent of this ui item.

        """
        self.ui_item.Create(parent, label=self._label)

        top, other = self.ui_item.GetBordersForSizer()
        self._bottom_border = other + self._extra_bottom_border
        self._sizer.AddSpacer(top + 5)
        self.ui_item.SetSizer(self._sizer)
        return self

    def __enter__(self):
        """Enter context manager."""
        return self

    def __exit__(self, *args):
        """Exit context manager."""
        self._finish()

    def _finish(self):
        self._sizer.AddSpacer(self._bottom_border)



import asyncio
import logging

from wxasync import WxAsyncApp

from demo.windows.main_window import MainWindow

_LOGGER = logging.getLogger(__name__)

if __name__ == "__main__":
    # os.environ["DEBUGGING"] = "1"
    logging.basicConfig(level=logging.DEBUG)
    loop = asyncio.get_event_loop()
    app = WxAsyncApp()
    main_window = MainWindow()
    main_window.Show()
    app.SetTopWindow(main_window)
    loop.run_until_complete(app.MainLoop())



"""Text entry widgets."""

# pylint: disable=invalid-name,no-self-use

import logging
from typing import Union, Optional

import wx

from aio_wx_widgets.binding import Bindable, Binding
from aio_wx_widgets.widgets.validators.validators import IntValidator

_LOGGER = logging.getLogger(__name__)


class Entry(Bindable):
    """A general textctrl for any input.

    Inherits from Bindable as it
    """

    def __init__(
        self,
        binding: Binding,
        label: Union[None, int, str] = None,
        wx_ctrl=None,
        validator: wx.Validator = None,
    ):

        super().__init__(binding)
        self.value: Optional[int] = None

        self._label = label

        if wx_ctrl is None:
            self._txt = wx.TextCtrl()
        else:
            self._txt = wx_ctrl

        self._validator = validator
        self.ui_item = self._txt

    def _set_ui_value(self, value):
        self._txt.SetValue(str(value))

    def _get_ui_value(self):
        return self._txt.GetValue()

    def __call__(self, parent):
        args = dict({"parent": parent})
        if self._validator:
            args["validator"] = self._validator
        if self._label:
            args["value"] = str(self._label)

        self._txt.Create(**args)
        self._txt.Bind(wx.EVT_TEXT, self._on_ui_change)
        self._make_binding()

        return self


class IntEntry(Entry):
    """Textentry which only allows numerical values."""

    def __init__(self, binding: "Binding", label: Union[None, int, str] = None):
        super().__init__(binding=binding, label=label, validator=IntValidator())


def text_ctrl(
    label=None, parent=None, multiline=False, validator=wx.DefaultValidator,
):
    """Create a wx text control."""
    txt = wx.TextCtrl()
    style = 0
    if multiline:
        style = style | wx.TE_MULTILINE

    def _create(parent):
        args = {"style": style, "validator": validator}
        if label:
            args["value"] = str(label)

        txt.Create(parent, **args)
        return txt

    if parent:
        return _create(parent)
    return _create



"""WX ui frames. A windows that holds a panel."""

import logging
from pathlib import Path
from typing import Optional

import wx

_LOGGER = logging.getLogger(__name__)


def _get_app_icon_32(img_path: Path) -> wx.Icon:
    app_icon_32 = wx.Icon()
    app_icon_32.CopyFromBitmap(wx.Bitmap(str(img_path), wx.BITMAP_TYPE_PNG))
    return app_icon_32


class DefaultFrame(wx.Frame):
    """A default frame for app windows.

    Always add a panel to this frame first.
    """

    def __init__(
        self,
        title,
        parent=None,
        size: wx.Size = None,
        style=None,
        icon_img: Optional[Path] = None,
    ):
        """Init.

        Args:
            title:
            parent:
            size:
            style:
        """
        kwargs = {}
        if style is not None:
            kwargs["style"] = style

        if size is None:
            size = wx.Size(800, 600)

        kwargs["size"] = size
        kwargs["title"] = title

        wx.Frame.__init__(self, parent, **kwargs)
        if icon_img:
            self.SetIcon(_get_app_icon_32(icon_img))

        self.Bind(wx.EVT_CLOSE, self._on_close)

    # pylint: disable=no-self-use
    def _on_close(self, evt):
        _LOGGER.debug("Frame is closed.")
        evt.Skip()



"""WX colors"""

import wx

MAGENTA = wx.Colour(255, 102, 0)
GREEN = wx.Colour(68, 170, 0)
RED = wx.Colour(212, 0, 0)
BLUE = wx.BLUE
DARK_GREY = wx.Colour(33, 33, 33)
LIGHT_GREY = wx.Colour(210, 210, 210)
WHITE = wx.Colour(255, 255, 255)



from __future__ import annotations

import logging

from aio_wx_widgets.binding import Binding
from aio_wx_widgets.panels.splitter_panel import TwoSplitterWindow
from aio_wx_widgets.sizers import AlignHorizontal, PanelMixin, SizerMixin
from aio_wx_widgets.widgets.button import AioButton
from aio_wx_widgets.widgets.grid import Grid, VERTICAL
from aio_wx_widgets.widgets.group import Group
from aio_wx_widgets.widgets.text import Text
from aio_wx_widgets.widgets.text_entry import IntEntry, Entry

_LOGGER = logging.getLogger(__name__)


class SplitterWindow(TwoSplitterWindow, PanelMixin, SizerMixin):
    def __init__(self, parent, controller, scrollable=False, window_one_width=250):
        super().__init__(
            parent, splitter_one_scrollable=False, splitter_two_scrollable=False
        )
        self._controller = controller
        self.splitter_window_two.add(Text("This is text"))
        self.splitter_window_two.add(IntEntry(self.bind("value_1")))

    @property
    def controller(self):
        return self._controller


class ViewThree(SplitterWindow):
    def __init__(self, parent, controller):
        self._controller = controller
        super().__init__(parent, controller)

    @property
    def controller(self):
        return self._controller

    def populate(self):
        """Populate this view."""
        self.splitter_window_one.add(IntEntry(binding=self.bind("value_1")))
        # Use a context manager for container types like a group or grid.
        # A group is a container with a label and a sizer inside. Inside
        # this sizer widgets, or other containers can be placed.
        with self.add(Group("A labelled container.")) as group:
            group.add(Text(text="A horizontal grid."))

            with group.add(Grid()) as grd:
                # the binding binds to an attribute defined in the controller
                # the weight determines how much space a specific item should consume
                # with respect to the other members of the container.
                grd.add(IntEntry(binding=self.bind("value_1")), weight=6, margin=3)
                grd.add(IntEntry(binding=self.bind("value_1")), weight=4, margin=3)
                grd.add(IntEntry(binding=self.bind("value_1")), weight=4, margin=3)

        with self.add(Group("Any text entry.")) as group:
            group.add(Entry(binding=Binding(self._controller, "a_string_value")))
            group.add(Entry(binding=Binding(self._controller, "a_string_value")))

        with self.add(Grid()) as grd:
            grd.add(AioButton("button one", self._set_value), weight=6)
            grd.add(AioButton("button two,self", self._set_value), weight=6)

        with self.add(Group("Nesting of grids")) as grp:
            with grp.add(Grid()) as grd:
                grd.add(
                    Text(text="Right aligned text"),
                    weight=3,
                    margin=5,
                    align_horizontal=AlignHorizontal.right,
                )
                with grd.add(Grid(VERTICAL), weight=6, margin=0) as vert_grid:
                    vert_grid.add(
                        AioButton("Left aligned button", self._set_value),
                        align_horizontal=AlignHorizontal.left,
                        margin=4,
                    )
                    vert_grid.add(AioButton("Set number entries.", self._set_value))
                    vert_grid.add(
                        Text(text="Center aligned text with a large margin."),
                        margin=(10, 10, 30, 5),
                        align_horizontal=AlignHorizontal.center,
                    )
                    vert_grid.add(
                        AioButton("right aligned button.", self._set_value),
                        weight=2,
                        align_horizontal=AlignHorizontal.right,
                    )

        self.add(AioButton("open other window", self._on_open_second_window))

    async def _on_open_second_window(self, evt):
        await self._controller.open_other_window()

    async def _on_open(self, evt):
        await self._controller.open_other_window()

    async def _set_value(self, evt):
        await self._controller.set_value()



"""Panel. Sits in a frame. Holds widgets."""
from __future__ import annotations

import logging

import wx

from aio_wx_widgets import type_annotations as T
from aio_wx_widgets.sizers import PanelMixin, SizerMixin

_LOGGER = logging.getLogger(__name__)


class SimplePanel(PanelMixin, SizerMixin):
    """A simple panel."""

    def __init__(self, parent, **kwargs):
        """Init."""
        self.ui_item = wx.Panel(parent)
        super().__init__()

        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.ui_item.SetSizer(self._sizer)

    @property
    def controller(self) -> T.BaseController:
        raise NotImplementedError()



"""All type annotations go here."""

from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from aio_wx_widgets.binding import Binding

    from aio_wx_widgets.controller import BaseController
    from asyncio import AbstractEventLoop

    try:
        import Protocol  # type: ignore
    except ImportError:
        from typing_extensions import Protocol

    assert BaseController
    assert AbstractEventLoop
    assert Binding

    class Choice(Protocol):
        """Typing protocol to be used for static type checking."""

        label: str
        value: Any



from invoke import task


@task
def prep_project(ctx):
    print(">> Creating virtual env")
    ctx.run("python -m venv venv")
    print(">> Updating pip")
    ctx.run("venv\\Scripts\\python.exe -m pip install pip -U")
    print(">> Installing flit")
    ctx.run("venv\\Scripts\\python.exe -m pip install flit")
    print(">> Installing flit dependencies")
    ctx.run("venv\\Scripts\\python.exe -m flit install")



"""Grid layout item."""

# pylint: disable=duplicate-code


import logging

import wx

from aio_wx_widgets.sizers import SizerMixin

_LOGGER = logging.getLogger(__name__)

VERTICAL = wx.VERTICAL
HORIZONTAL = wx.HORIZONTAL


class Grid(SizerMixin):
    """A grid component.

    Basically a boxsizer with some defaults.
    """

    default_sizer_margin = 0

    def __init__(self, orientation=HORIZONTAL):
        """Init."""
        self.ui_item = self._sizer = wx.BoxSizer(orient=orientation)
        self._parent = None

    def __call__(self, parent):
        # setting a specific parent.
        self._parent = parent
        return self

    def __enter__(self):
        """Enter context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context manager."""



"""A dropdown selection widget."""

from __future__ import annotations
import logging
from typing import Callable, Optional, Sequence

import wx
from aio_wx_widgets import type_annotations as T
from aio_wx_widgets.binding import Bindable

_LOGGER = logging.getLogger(__name__)


class Select(Bindable):
    """Dropdown selection widget.

    Args:
        choices: Used to populate the dropdown.
        on_select_call_back: called when selection is made.
        default:
    """

    def __init__(
        self,
        choices: Sequence[T.Choice],
        on_select_callback: Optional[Callable[[T.Choice], None]] = None,
        binding: Optional[T.Binding] = None,
    ):
        super().__init__(binding)
        self.choices = choices
        self._on_select_callback = on_select_callback
        self._selected_item = None

        self.ui_item = wx.Choice()
        self.ui_item.Bind(wx.EVT_CHOICE, self._on_choice)
        self.ui_item.Bind(wx.EVT_MOUSEWHEEL, self._on_mouse_wheel)

    def _on_mouse_wheel(self, evt):
        """Capturing as I want to disable selection when scrolling"""

    def init(self, parent):
        """Initialize the component."""
        self.ui_item.Create(parent, choices=[choice.label for choice in self.choices])

    def _set_ui_value(self, value: T.Choice):
        try:
            _idx = next(
                (i for i, val in enumerate(self.choices) if val.value == value.value)
            )
        except StopIteration:
            _LOGGER.error("Default product not available")
        else:
            self.ui_item.Select(_idx)

    def _on_choice(self, event):
        _idx = self.ui_item.GetSelection()
        _LOGGER.debug("Selected idx: %s", _idx)
        self._selected_item = self.choices[_idx]
        self._on_ui_change()
        if self._on_select_callback:
            self._on_select_callback(self._selected_item)

    def _get_ui_value(self):
        return self._selected_item

    def __call__(self, parent):
        self.init(parent)
        self._make_binding()

        return self
